from typing import Optional
from abc import ABC, abstractmethod
from langchain_core.tools import StructuredTool
from langchain_core.pydantic_v1 import BaseModel

from motleycrew.agents.abstract_parent import MotleyAgentAbstractParent
from motleycrew.common.exceptions import InvalidOutput
from motleycrew.tools import MotleyTool


class MotleyOutputHandler(MotleyTool, ABC):
    _name: str = "output_handler"
    """Name of the output handler tool."""

    _description: str = "Output handler. ONLY RETURN THE FINAL RESULT USING THIS TOOL!"
    """Description of the output handler tool."""

    _args_schema: Optional[BaseModel] = None
    """Pydantic schema for the arguments of the output handler tool.
    Inferred from the `handle_output` method if not provided."""

    _exceptions_to_handle: tuple[Exception] = (InvalidOutput,)
    """Exceptions that should be returned to the agent when raised in the `handle_output` method."""

    def __init__(self):
        langchain_tool = self._create_langchain_tool()
        super().__init__(langchain_tool)

        self.agent: Optional[MotleyAgentAbstractParent] = None
        self.agent_input: Optional[dict] = None

    @property
    def exceptions_to_handle(self):
        return self._exceptions_to_handle

    def _create_langchain_tool(self):
        return StructuredTool.from_function(
            name=self._name,
            description=self._description,
            args_schema=self._args_schema,
            func=self.handle_output,
        )

    @abstractmethod
    def handle_output(self, *args, **kwargs):
        pass



from datetime import datetime
from pathlib import Path
from typing import Tuple, Optional

from motleycrew.common.utils import ensure_module_is_installed

try:
    from selenium import webdriver
    from selenium.webdriver.chrome.service import Service
except ImportError:
    webdriver = None
    Service = None

from langchain.tools import Tool
from langchain_core.pydantic_v1 import BaseModel, Field

from motleycrew.tools import MotleyTool
from motleycrew.common import logger


class HTMLRenderer:
    def __init__(
        self,
        work_dir: str,
        executable_path: str | None = None,
        headless: bool = True,
        window_size: Optional[Tuple[int, int]] = None,
    ):
        """Helper for rendering HTML code as an image"""
        ensure_module_is_installed(
            "selenium",
            "see documentation: https://pypi.org/project/selenium/, ChromeDriver is also required",
        )

        self.work_dir = Path(work_dir).resolve()
        self.html_dir = self.work_dir / "html"
        self.images_dir = self.work_dir / "images"

        self.options = webdriver.ChromeOptions()
        if headless:
            self.options.add_argument("--headless")
        self.service = Service(executable_path=executable_path)

        self.window_size = window_size

    def render_image(self, html: str, file_name: str | None = None):
        """Create image with png extension from html code

        Args:
            html (str): html code for rendering image
            file_name (str): file name with not extension
        Returns:
            file path to created image
        """
        logger.info("Trying to render image from HTML code")
        html_path, image_path = self.build_save_file_paths(file_name)
        browser = webdriver.Chrome(options=self.options, service=self.service)
        try:
            if self.window_size:
                logger.info("Setting window size to {}".format(self.window_size))
                browser.set_window_size(*self.window_size)

            url = "data:text/html;charset=utf-8,{}".format(html)
            browser.get(url)

            logger.info("Taking screenshot")
            is_created_img = browser.get_screenshot_as_file(image_path)
        finally:
            browser.close()
            browser.quit()

        if not is_created_img:
            logger.error("Failed to render image from HTML code {}".format(image_path))
            return "Failed to render image from HTML code"

        with open(html_path, "w") as f:
            f.write(html)
        logger.info("Saved the HTML code to {}".format(html_path))
        logger.info("Saved the rendered HTML screenshot to {}".format(image_path))

        return image_path

    def build_save_file_paths(self, file_name: str | None = None) -> Tuple[str, str]:
        """Builds paths to html and image files

        Args:
            file_name (str): file name with not extension

        Returns:
            tuple[str, str]: html file path and image file path
        """

        # check exists dirs:
        for _dir in (self.work_dir, self.html_dir, self.images_dir):
            if not _dir.exists():
                _dir.mkdir(parents=True)

        file_name = file_name or datetime.now().strftime("%Y_%m_%d__%H_%M")
        html_path = self.html_dir / "{}.html".format(file_name)
        image_path = self.images_dir / "{}.png".format(file_name)

        return str(html_path), str(image_path)


class HTMLRenderTool(MotleyTool):

    def __init__(
        self,
        work_dir: str,
        executable_path: str | None = None,
        headless: bool = True,
        window_size: Optional[Tuple[int, int]] = None,
    ):
        """Tool for rendering HTML as image

        Args:
            work_dir (str): Directory for saving images and html files
        """
        renderer = HTMLRenderer(
            work_dir=work_dir,
            executable_path=executable_path,
            headless=headless,
            window_size=window_size,
        )
        langchain_tool = create_render_tool(renderer)
        super(HTMLRenderTool, self).__init__(langchain_tool)


class HTMLRenderToolInput(BaseModel):
    """Input for the HTMLRenderTool.

    Attributes:
        html (str):
    """

    html: str = Field(description="HTML code for rendering")


def create_render_tool(renderer: HTMLRenderer):
    """Create langchain tool from HTMLRenderer.render_image method

    Returns:
        Tool:
    """
    return Tool.from_function(
        func=renderer.render_image,
        name="HTML rendering tool",
        description="A tool for rendering HTML code as an image",
        args_schema=HTMLRenderToolInput,
    )



""" Module description """

from typing import Any, Optional, Sequence

from langchain.agents import AgentExecutor
from langchain_core.chat_history import InMemoryChatMessageHistory
from langchain_core.runnables import RunnableConfig
from langchain_core.runnables.history import RunnableWithMessageHistory, GetSessionHistoryCallable

from motleycrew.agents.mixins import LangchainOutputHandlingAgentMixin
from motleycrew.agents.parent import MotleyAgentParent
from motleycrew.common import MotleyAgentFactory
from motleycrew.common import MotleySupportedTool, logger
from motleycrew.tracking import add_default_callbacks_to_langchain_config


class LangchainMotleyAgent(MotleyAgentParent, LangchainOutputHandlingAgentMixin):
    def __init__(
        self,
        prompt_prefix: str | None = None,
        description: str | None = None,
        name: str | None = None,
        agent_factory: MotleyAgentFactory[AgentExecutor] | None = None,
        tools: Sequence[MotleySupportedTool] | None = None,
        output_handler: MotleySupportedTool | None = None,
        verbose: bool = False,
        chat_history: bool | GetSessionHistoryCallable = True,
    ):
        """Description

        Args:
            prompt_prefix (:obj:`str`, optional):
            description (:obj:`str`, optional):
            name (:obj:`str`, optional):
            agent_factory (:obj:`MotleyAgentFactory`, optional):
            tools (:obj:`Sequence[MotleySupportedTool]`, optional):
            output_handler (:obj:`MotleySupportedTool`, optional):
            verbose (bool):
            chat_history (:obj:`bool`, :obj:`GetSessionHistoryCallable`):
            Whether to use chat history or not. If `True`, uses `InMemoryChatMessageHistory`.
            If a callable is passed, it is used to get the chat history by session_id.
            See Langchain `RunnableWithMessageHistory` get_session_history param for more details.
        """
        super().__init__(
            prompt_prefix=prompt_prefix,
            description=description,
            name=name,
            agent_factory=agent_factory,
            tools=tools,
            output_handler=output_handler,
            verbose=verbose,
        )

        self._agent_finish_blocker_tool = self._create_agent_finish_blocker_tool()

        if chat_history is True:
            chat_history = InMemoryChatMessageHistory()
            self.get_session_history_callable = lambda _: chat_history
        else:
            self.get_session_history_callable = chat_history

    def materialize(self):
        """Materialize the agent and wrap it in RunnableWithMessageHistory if needed."""
        if self.is_materialized:
            return

        super().materialize()
        assert isinstance(self._agent, AgentExecutor)

        if self.output_handler:
            self._agent.tools += [self._agent_finish_blocker_tool]

            object.__setattr__(
                self._agent.agent, "plan", self.agent_plan_decorator(self._agent.agent.plan)
            )

            object.__setattr__(
                self._agent,
                "_take_next_step",
                self.take_next_step_decorator(self._agent._take_next_step),
            )

            prepared_output_handler = None
            for tool in self.agent.tools:
                if tool.name == self.output_handler.name:
                    prepared_output_handler = tool

            object.__setattr__(
                prepared_output_handler,
                "_run",
                self._run_tool_direct_decorator(prepared_output_handler._run),
            )

            object.__setattr__(
                prepared_output_handler,
                "run",
                self.run_tool_direct_decorator(prepared_output_handler.run),
            )

        if self.get_session_history_callable:
            logger.info("Wrapping agent in RunnableWithMessageHistory")

            if isinstance(self._agent, RunnableWithMessageHistory):
                return
            self._agent = RunnableWithMessageHistory(
                runnable=self._agent,
                get_session_history=self.get_session_history_callable,
                input_messages_key="input",
                history_messages_key="chat_history",
            )

    def invoke(
        self,
        input: dict,
        config: Optional[RunnableConfig] = None,
        **kwargs: Any,
    ) -> Any:
        """Description

        Args:
            input (dict):
            config (:obj:`RunnableConfig`, optional):
            **kwargs:

        Returns:

        """
        prompt = self.prepare_for_invocation(input=input)

        config = add_default_callbacks_to_langchain_config(config)
        if self.get_session_history_callable:
            config["configurable"] = config.get("configurable") or {}
            config["configurable"]["session_id"] = (
                config["configurable"].get("session_id") or "default"
            )

        output = self.agent.invoke({"input": prompt}, config, **kwargs)
        output = output.get("output")
        return output

    @staticmethod
    def from_agent(
        agent: AgentExecutor,
        goal: str,
        description: str | None = None,
        tools: Sequence[MotleySupportedTool] | None = None,
        verbose: bool = False,
    ) -> "LangchainMotleyAgent":
        """Description

        Args:
            agent (AgentExecutor):
            goal (str):
            description (:obj:`str`, optional)
            tools(:obj:`Sequence[MotleySupportedTool]`, optional):
            verbose (bool):

        Returns:
            LangchainMotleyAgent
        """
        # TODO: do we really need to unite the tools implicitly like this?
        # TODO: confused users might pass tools both ways at the same time
        # TODO: and we will silently unite them, which can have side effects (e.g. doubled tools)
        # TODO: besides, getting tools can be tricky for other frameworks (e.g. LlamaIndex)
        if tools or agent.tools:
            tools = list(tools or []) + list(agent.tools or [])

        wrapped_agent = LangchainMotleyAgent(
            prompt_prefix=goal, description=description, tools=tools, verbose=verbose
        )
        wrapped_agent._agent = agent
        return wrapped_agent



""" Module description """

import uuid
from typing import Any, Optional, Sequence

try:
    from llama_index.core.agent import AgentRunner
    from llama_index.core.chat_engine.types import ChatResponseMode
    from llama_index.core.agent.types import TaskStep, TaskStepOutput
    from llama_index.core.chat_engine.types import AgentChatResponse
except ImportError:
    AgentRunner = None
    ChatResponseMode = None
    TaskStep = None
    TaskStepOutput = None
    AgentChatResponse = None

from langchain_core.runnables import RunnableConfig

from motleycrew.agents.parent import MotleyAgentParent, DirectOutput
from motleycrew.common import MotleySupportedTool
from motleycrew.common import MotleyAgentFactory
from motleycrew.common.utils import ensure_module_is_installed


class LlamaIndexMotleyAgent(MotleyAgentParent):

    def __init__(
        self,
        prompt_prefix: str | None = None,
        description: str | None = None,
        name: str | None = None,
        agent_factory: MotleyAgentFactory[AgentRunner] | None = None,
        tools: Sequence[MotleySupportedTool] | None = None,
        output_handler: MotleySupportedTool | None = None,
        verbose: bool = False,
    ):
        """Description

        Args:
            prompt_prefix (:obj:`str`, optional):
            description (:obj:`str`, optional):
            name (:obj:`str`, optional):
            agent_factory (:obj:`MotleyAgentFactory`, optional):
            tools (:obj:`Sequence[MotleySupportedTool]`, optional):
            verbose (:obj:`bool`, optional):
        """
        super().__init__(
            description=description,
            prompt_prefix=prompt_prefix,
            name=name,
            agent_factory=agent_factory,
            tools=tools,
            output_handler=output_handler,
            verbose=verbose,
        )

    def run_step_decorator(self):
        """Decorator for inclusion in the call chain of the agent, the output handler tool"""
        ensure_module_is_installed("llama_index")

        def decorator(func):
            def wrapper(
                task_id: str,
                step: Optional[TaskStep] = None,
                input: Optional[str] = None,
                mode: ChatResponseMode = ChatResponseMode.WAIT,
                **kwargs: Any,
            ):

                try:
                    cur_step_output = func(task_id, step, input, mode, **kwargs)
                except DirectOutput as output_exc:
                    self.direct_output = output_exc
                    output = AgentChatResponse(str(output_exc.output))
                    task_step = TaskStep(task_id=task_id, step_id=str(uuid.uuid4()))
                    cur_step_output = TaskStepOutput(
                        output=output, is_last=True, next_steps=[], task_step=task_step
                    )
                    return cur_step_output

                if self.output_handler is None:
                    return cur_step_output

                if cur_step_output.is_last:
                    cur_step_output.is_last = False
                    task_id = cur_step_output.task_step.task_id
                    output_task_step = TaskStep(
                        task_id=task_id,
                        step_id=str(uuid.uuid4()),
                        input="You must call the `{}` tool to return the output.".format(
                            self.output_handler.name
                        ),
                    )

                    cur_step_output.next_steps.append(output_task_step)

                    step_queue = self._agent.state.get_step_queue(task_id)
                    step_queue.extend(cur_step_output.next_steps)

                return cur_step_output

            return wrapper

        return decorator

    def materialize(self):
        super(LlamaIndexMotleyAgent, self).materialize()
        self._agent._run_step = self.run_step_decorator()(self._agent._run_step)

    def invoke(
        self,
        input: dict,
        config: Optional[RunnableConfig] = None,
        **kwargs: Any,
    ) -> Any:
        """Description

        Args:
            input (dict):
            config (:obj:`RunnableConfig`, optional):
            **kwargs:

        Returns:
            Any:
        """
        prompt = self.prepare_for_invocation(input=input)

        output = self.agent.chat(prompt)

        if self.output_handler:
            return self.direct_output.output

        return output.response

    @staticmethod
    def from_agent(
        agent: AgentRunner,
        description: Optional[str] = None,
        prompt_prefix: Optional[str] = None,
        tools: Sequence[MotleySupportedTool] | None = None,
        verbose: bool = False,
    ) -> "LlamaIndexMotleyAgent":
        """Description

        Args:
            agent (AgentRunner):
            description (:obj:`str`, optional):
            prompt_prefix (:obj:`str`, optional):
            tools (:obj:`Sequence[MotleySupportedTool]`, optional):
            verbose (:obj:`bool`, optional):

        Returns:
            LlamaIndexMotleyAgent:
        """
        ensure_module_is_installed("llama_index")
        wrapped_agent = LlamaIndexMotleyAgent(
            description=description, prompt_prefix=prompt_prefix, tools=tools, verbose=verbose
        )
        wrapped_agent._agent = agent
        return wrapped_agent



from typing import Any, Optional, Callable, Union, Dict, List, Tuple

from langchain_core.agents import AgentFinish, AgentAction
from langchain_core.callbacks import CallbackManagerForChainRun
from langchain_core.messages import AIMessage
from langchain_core.tools import BaseTool, Tool

from motleycrew.agents.parent import DirectOutput
from motleycrew.tools import MotleyTool


class LangchainOutputHandlingAgentMixin:
    """A mixin for Langchain-based agents that support output handlers."""

    output_handler: Optional[MotleyTool] = None
    _agent_finish_blocker_tool: Optional[MotleyTool] = None

    def _create_agent_finish_blocker_tool(self) -> BaseTool:
        """Create a tool that will force the agent to retry if it attempts to return the output
        bypassing the output handler.
        """

        def create_agent_finish_blocking_message(input: Any = None) -> str:
            return (
                f"You must call the `{self.output_handler.name}` tool to return the final output.\n"
            )

        return Tool.from_function(
            name="agent_finish_blocker",
            description="",
            func=create_agent_finish_blocking_message,
        )

    def _is_blocker_action(self, action: AgentAction) -> bool:
        """Checks whether the action of the response blocking tool"""
        return bool(
            isinstance(action, AgentAction) and action.tool == self._agent_finish_blocker_tool.name
        )

    def agent_plan_decorator(self, func: Callable):
        """Decorator for Agent.plan() method that intercepts AgentFinish events"""

        def wrapper(
            intermediate_steps: List[Tuple[AgentAction, str]],
            callbacks: "Callbacks" = None,
            **kwargs: Any,
        ) -> Union[AgentAction, AgentFinish]:
            additional_notes = []

            if self.output_handler:
                to_remove_steps = []
                for intermediate_step in intermediate_steps:
                    action, action_output = intermediate_step
                    if self._is_blocker_action(action):
                        # Add the interaction telling the LLM that it must use the output handler
                        additional_notes.append(("ai", action.tool_input))
                        additional_notes.append(("user", action_output))
                        to_remove_steps.append(intermediate_step)

                for to_remove_step in to_remove_steps:
                    intermediate_steps.remove(to_remove_step)

            if additional_notes:
                kwargs["additional_notes"] = additional_notes

            step = func(intermediate_steps, callbacks, **kwargs)

            if not isinstance(step, AgentFinish):
                return step

            if self.output_handler is not None:
                return AgentAction(
                    tool=self._agent_finish_blocker_tool.name,
                    tool_input=step.log,
                    log="\nDetected AgentFinish, blocking it to force output via output handler.\n",
                )
            return step

        return wrapper

    def take_next_step_decorator(self, func: Callable):
        """
        Decorator for AgentExecutor._take_next_step() method that catches DirectOutput exceptions.
        """

        def wrapper(
            name_to_tool_map: Dict[str, BaseTool],
            color_mapping: Dict[str, str],
            inputs: Dict[str, str],
            intermediate_steps: List[Tuple[AgentAction, str]],
            run_manager: Optional[CallbackManagerForChainRun] = None,
        ) -> Union[AgentFinish, List[Tuple[AgentAction, str]]]:

            try:
                step = func(
                    name_to_tool_map, color_mapping, inputs, intermediate_steps, run_manager
                )
            except DirectOutput as direct_ex:
                message = str(direct_ex.output)
                return AgentFinish(
                    return_values={"output": direct_ex.output},
                    messages=[AIMessage(content=message)],
                    log=message,
                )
            return step

        return wrapper

    def _run_tool_direct_decorator(self, func: Callable):
        """Decorator of the tool's _run method, for intercepting a DirectOutput exception"""

        def wrapper(*args, **kwargs):
            try:
                result = func(*args, **kwargs)
            except DirectOutput as direct_exc:
                return direct_exc
            return result

        return wrapper

    def run_tool_direct_decorator(self, func: Callable):
        """Decorator of the tool's run method, for intercepting a DirectOutput exception"""

        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            if isinstance(result, DirectOutput):
                raise result
            return result

        return wrapper
